package edu.bu.signstream.grepresentation.fields.glossField.dialog.dai;

import edu.bu.signstream.common.files.SSResources;
import edu.bu.signstream.common.util.ErrorMessages;
import edu.bu.signstream.grepresentation.requestsHandler.SS3Singleton;
import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.Time;
import edu.bu.signstream.ui.SS3SignStreamApplication;

import javax.swing.*;
import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.concurrent.CancellationException;
import java.util.logging.Level;

/**
 * This class will take a video file and splice it based on the start and end
 */
public class VideoSplicer {
    private final String ffmpegPath; // Path to ffmpeg executable stored in SignStream
    private final String videoFileResultName; // The name of the spliced video
    private Time startTime;
    private Time endTime;
    private File videoFile; // Video file to splice
    private ProcessBuilder processBuilder;
    private final String splicedVideoPath; // Where to store the spliced video with the name of the spliced video appended to the end

    /**
     * @param startTime         Start time of the video to splice
     * @param endTime          End time of the video to splice
     * @param videoFile       Video file to splice
     * @param videoFileResultName Name of the spliced video
     */
    public VideoSplicer(Time startTime, Time endTime, File videoFile, String videoFileResultName) {
        this.ffmpegPath = SSResources.ffmpegPath;
        this.startTime = startTime;
        this.endTime = endTime;
        this.videoFile = videoFile;
        this.splicedVideoPath = SSResources.ffmpegOutputDirectoryPath + videoFileResultName;
        this.videoFileResultName = videoFileResultName;
    }

    /**
     * See <a href="https://ffmpeg.org/ffmpeg.html">ffmpeg documentation</a> for more information on the ffmpeg command
     * @return True if the splicer was initialized successfully, false otherwise
     */
    public boolean initializeSplicer() {
        Duration totalDuration = this.endTime.subtract(this.startTime);

        double startTimeInSeconds = this.startTime.value() / 1000.0;
        double totalDurationInSeconds = totalDuration.value() / 1000.0;

        // https://ffmpeg.org/ffmpeg.html

        String[] command = {this.ffmpegPath, "-y", "-ss", String.valueOf(startTimeInSeconds), "-i", this.videoFile.getAbsolutePath(),
                "-t", String.valueOf(totalDurationInSeconds), "-c", "copy", "-f", "mp4", this.videoFileResultName}; // -y to overwrite file if it exists
        this.processBuilder = new ProcessBuilder(command);
        this.processBuilder.redirectErrorStream(true); // Redirect error stream to output stream

        File tempDirectory = new File(SSResources.ffmpegOutputDirectoryPath); // Where the spliced video will be stored
        try {
            if (!tempDirectory.exists()) { // Create temp directory if it doesn't exist
                if (tempDirectory.mkdir() || tempDirectory.mkdirs()) {
                    SS3SignStreamApplication.logger.log(Level.INFO, "Created temp directory");
                } else {
                    String warningMessage = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_TEMP_DIRECTORY);
                    String warningMessageTitle = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_TEMP_DIRECTORY_TITLE);

                    JOptionPane.showMessageDialog(SS3Singleton.getSignStreamApplication().getSignStreamSegmentPanel(), warningMessage, warningMessageTitle, JOptionPane.ERROR_MESSAGE);
                    SS3SignStreamApplication.logger.log(Level.SEVERE, "Failed to create temp directory");

                    return false;
                }
            } else {
                // Delete spliced video from temp directory if it exists
                File videoFileToDelete = this.getSplicedVideoFile();

                if (videoFileToDelete.exists()) {
                    if (this.deleteSpecificVideoFile(videoFileToDelete)) {
                        return false;
                    }
                }

            }
        } catch (SecurityException e) {
            String warningMessage = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_SPLICING_ERROR_SECURITY_EXCEPTION);
            String warningMessageTitle = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_TEMP_DIRECTORY_TITLE);

            JOptionPane.showMessageDialog(SS3Singleton.getSignStreamApplication().getSignStreamSegmentPanel(), warningMessage, warningMessageTitle, JOptionPane.ERROR_MESSAGE);
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Failed to create temp directory");
            return false;
        }

        this.processBuilder.directory(new File(SSResources.ffmpegOutputDirectoryPath)); // Set the working directory to the temp directory
        return true;
    }

    private boolean deleteSpecificVideoFile(File videoFileToDelete) {
        if (!videoFileToDelete.delete()) {
            String warningMessage = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_TEMP_DIRECTORY_DELETION_ERROR);
            String warningMessageTitle = SS3Singleton.getErrorMessages().getErrorMessageString(ErrorMessages.VIDEO_SPLICER_ERROR_TEMP_DIRECTORY_TITLE);

            JOptionPane.showMessageDialog(SS3Singleton.getSignStreamApplication().getSignStreamSegmentPanel(), warningMessage, warningMessageTitle, JOptionPane.ERROR_MESSAGE);
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Failed to delete file in temp directory");
            return true;
        }
        return false;
    }

    private boolean deleteAllFiles(File[] files) {
        if (files != null) {
            for (File file : files) {
                if (this.deleteSpecificVideoFile(file)) {
                    return false;
                }
            }
        }

        return true;
    }

    public File getSplicedVideoFile() {
        return new File(this.splicedVideoPath);
    }

    /**
     * Start to splice the video
     * @return True if the video was spliced successfully, false otherwise
     */
    public boolean spliceVideo() {
        int exitValue;
        try {
            Process process = this.processBuilder.start(); // Start splicing video
            exitValue = process.waitFor(); // Wait for splicing to finish

            BufferedReader reader = new BufferedReader(new InputStreamReader(process.getInputStream()));
            String line;

            while ((line = reader.readLine()) != null) {
                System.out.println(line); // Read the output
            }

            reader.close();


            return exitValue == 0;
        } catch (IOException | InterruptedException | CancellationException e) {
            SS3SignStreamApplication.logger.log(Level.SEVERE, "Error splicing video and/or it was cancelled", e);

            return false;
        }
    }

    public String getFfmpegPath() {
        return this.ffmpegPath;
    }

    public Time getStartTime() {
        return this.startTime;
    }

    public void setStartTime(Time startTime) {
        this.startTime = startTime;
    }

    public Time getEndTime() {
        return this.endTime;
    }

    public void setEndTime(Time endTime) {
        this.endTime = endTime;
    }

    public File getVideoFile() {
        return this.videoFile;
    }

    public void setVideoFile(File videoFile) {
        this.videoFile = videoFile;
    }

    public ProcessBuilder getProcessBuilder() {
        return this.processBuilder;
    }

    public void setProcessBuilder(ProcessBuilder processBuilder) {
        this.processBuilder = processBuilder;
    }

}
