package edu.bu.signstream.media.gstreamer;


import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.time.TimeUnit;
import edu.bu.signstream.media.Audio;
import edu.bu.signstream.media.ui.AudioView;
import edu.bu.signstream.ui.frame.SSFrame;
import org.freedesktop.gstreamer.*;
import org.freedesktop.gstreamer.query.PositionQuery;

import javax.swing.*;
import java.net.URI;

public class GstAudio implements GstMedia, Audio {
    private static final String FORMAT = "audio/x-raw";
    private static final String ID_PREFIX = "Audio";

    private static int ID = 0;
    private final String id;

    private final URI uri;
    private final int order;

    private int rate;
    private Duration duration;

    private TimeRange range;

    private final GstMediaPlayer player;

    private final Element decode;
    private final Element convert;
    private final Element resample;
    private final Element scaletempo;
    private final Element sink;

    private AudioView view;

    public GstAudio(URI uri, int order, GstMediaPlayer player) {
        this.player = player;
        this.uri = uri;
        this.order = order;
        this.id = ID_PREFIX + ID;

        System.out.println("Probing audio " + uri.toString());
        var caps = GstProbe.probe(uri, FORMAT);
        System.out.println("Probing finished!");

        rate = (int)caps.get("rate");
        var unit = new TimeUnit(id, "SAMPLES", TimeUnit.Type.FRAME, rate);

        duration = ((Duration)caps.get("duration")).convert(unit);
        range = TimeRange.ABSOLUTE.subrange(id, unit, Duration.ZERO, duration, 100);

        // Set up the elements for the real pipeline
        decode = ElementFactory.make("uridecodebin", "decode" + id);
        decode.set("uri", uri.toString());
        decode.connect(new Element.PAD_ADDED() {
            @Override
            public void padAdded(Element e, Pad pad) {
				System.out.println(uri + ": AUDIO " + pad.getName() + " [" + pad.getTypeName() + "]");
				var caps = pad.getCurrentCaps().getStructure(0);
                System.out.println(caps);

                if (caps.getName().equals(FORMAT)) {
                    try {
                        System.out.println("LINKED AUDIO");
                        pad.link(convert.getStaticPad("sink"));
                    } catch (PadLinkException ex) {
                        ex.printStackTrace();
                    }
                }
            }
        });
        convert = ElementFactory.make("audioconvert", "convert" + id);
        resample = ElementFactory.make("audioresample", "resample" + id);
        scaletempo = ElementFactory.make("scaletempo", "scaletempo" + id);

        sink = ElementFactory.make("autoaudiosink", "sink" + id);

        player.pipeline.addMany(decode, convert, resample, scaletempo, sink);
        convert.link(resample);
        resample.link(scaletempo);
        scaletempo.link(sink);

        System.out.println("Done loading audio " + id);

        ID++;
    }

    String id() {
        return id;
    }

    @Override
    public URI uri() {
        return uri;
    }
    
    @Override
	public int order() {
    	return order;
    }

    @Override
    public int rate() {
        return rate;
    }

    @Override
    public Duration duration() {
        return duration;
    }

    @Override
    public TimeRange range() {
        return range;
    }

    @Override
    public void range(TimeRange range) {
        this.range = range;
    }

    @Override
    public SSFrame view() { return view; }

    @Override
    public void display() {
        view = new AudioView(player, this, () -> this.view = null);
    }

    @Override
    public void hide() {
        if (view != null)
            view.dispose();
        view = null;
    }

    @Override
    public boolean displayed() {
        return view != null;
    }

    @Override
    public long pos() {
        if (!Thread.currentThread().getName().contains("gstreamer service thread"))
            throw new AssertionError("Call to GstAudio.pos() from outside of GStreamer service thread");

        var q = new PositionQuery(Format.TIME);
        sink.query(q);
        return q.getPosition();
    }

    @Override
    public Element sink() {
        return sink;
    }

    @Override
    public void close() {
        if (!Thread.currentThread().getName().contains("gstreamer service thread"))
            throw new AssertionError("Call to GstAudio.close() from outside of GStreamer service thread");

        player.pipeline.removeMany(decode, convert, resample, scaletempo, sink);
    }
}
