package edu.bu.signstream.media.ui;

import edu.bu.signstream.common.SSMath;
import edu.bu.signstream.media.ui.icon.SizeLargeIconLight;
import edu.bu.signstream.media.ui.icon.SizeSmallIconLight;
import edu.bu.signstream.time.*;
import edu.bu.signstream.grepresentation.requestsHandler.SS3Singleton;
import edu.bu.signstream.media.Audio;
import edu.bu.signstream.media.MediaPlayer;
import edu.bu.signstream.media.gstreamer.GstAudioReader;
import edu.bu.signstream.media.ui.color.ColorMap;
import edu.bu.signstream.media.ui.color.Gradient;
import edu.bu.signstream.media.ui.painter.*;

import javax.swing.*;
import javax.swing.event.ChangeListener;
import java.awt.*;
import java.util.ArrayList;

public class AudioView extends MediaView {
    private final WaveformPainter waveform;
    private final SpectrogramPainter spectrogram;
    private final PitchPainter pitch;

    private final JPanel view = new JPanel();
    private final JPanel waveformView;
    private final JPanel spectrogramView;
    private final JPanel emptyView;

    private final ToggleButton resize = new ToggleButton(b -> b ? new SizeLargeIconLight(true) : new SizeSmallIconLight(true), false);
    private final JCheckBox waveformToggle = new JCheckBox("Waveform");
    private final JCheckBox spectrogramToggle = new JCheckBox("Spectrogram");
    private final JCheckBox pitchToggle = new JCheckBox("Pitch");

    private final Dimension largerSize;
    private final Dimension smallerSize;

    private static final ColorMap colorsMain = Gradient.of(Color.BLACK, Color.BLUE);
    private static final ColorMap colorsAccent = Gradient.of(Color.BLACK, Color.ORANGE);

    private static final int ZOOM_MIN = 0;
    private static final int ZOOM_MAX = 1000000;
    private static final Duration WINDOW_MIN = Duration.sec(0.5);
    private static final Duration WINDOW_MAX = Duration.sec(20);
    private static final Duration WINDOW_DEFAULT = Duration.sec(5);
    private final JSlider zoom = new JSlider(0, ZOOM_MAX);

    @Override
    protected JComponent view() { return view; }

    @Override
    protected JPanel controls(JPanel panel) {
        waveformToggle.setFocusable(false);
        spectrogramToggle.setFocusable(false);
        pitchToggle.setFocusable(false);
        zoom.setFocusable(false);
        zoom.setMajorTickSpacing(SLIDER_MAX / 5);
        zoom.setMinorTickSpacing(SLIDER_MAX / 20);
        zoom.setPaintTicks(true);

        panel.add(resize);
        panel.add(waveformToggle);
        panel.add(spectrogramToggle);
        panel.add(pitchToggle);
        panel.add(zoom);

        panel.setBackground(Color.WHITE);
        return panel;
    }

    @Override
    protected TimeRange defaultTimeRange() { return SS3Singleton.getVideoTimeRange(); }

    @Override
    protected void timeChanged(Time t, double fr) {
        super.timeChanged(t, fr);
        waveform.seek(fr);
        spectrogram.seek(fr);
        pitch.seek(fr);
        view.repaint();
    }

    public AudioView(MediaPlayer player, Audio audio, Runnable onClose) {
        super(player, audio, onClose);

        var loading = new JDialog(null, "Loading...", Dialog.ModalityType.MODELESS);
        var loadingPanel = new JPanel();
        loadingPanel.setBorder(BorderFactory.createEmptyBorder(10, 5, 10, 5));
        loadingPanel.add(new JLabel("Analyzing audio data. Please wait."));
        loading.add(loadingPanel);
        loading.pack();
        loading.setLocationRelativeTo(this);
        loading.setVisible(true);
        // Force the loading dialog to paint immediately
        var root = loading.getRootPane();
        var bounds = root.getBounds();
        root.paintImmediately(0, 0, bounds.width, bounds.height);

        var start = System.currentTimeMillis();
        var samples = GstAudioReader.read(audio.uri());
        var end = System.currentTimeMillis();
        System.out.println("Read " + samples.length + " samples at " + GstAudioReader.RATE + "Hz in " + ((end - start) / 1000.0) + "s");

        waveform = new WaveformPainter(samples, colorsMain, media.range(), WINDOW_DEFAULT);
        spectrogram = new SpectrogramPainter(samples, colorsMain, media.range(), WINDOW_DEFAULT);
        pitch = new PitchPainter(samples, colorsAccent, media.range(), WINDOW_DEFAULT);

        loading.dispose();

        waveformView = new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                var sz = this.getSize();
                g.setColor(Color.BLACK);
                g.fillRect(0, 0, sz.width, sz.height);

                waveform.paint(this, g);
                waveform.paintBar(this, g);
                waveform.paintCursor(this, g, Color.WHITE);
            }
        };

        spectrogramView = new JPanel() {
            protected void paintComponent(Graphics g) {
                var sz = this.getSize();
                g.setColor(Color.BLACK);
                g.fillRect(0, 0, sz.width, sz.height);

                var text = new ArrayList<String>();
                if (spectrogramToggle.isSelected()) {
                    spectrogram.paint(this, g);
                    text.addAll(spectrogram.text(this));
                }
                if (pitchToggle.isSelected()) {
                    pitch.paint(this, g);
                    pitch.paintNearest(this, g, Color.RED);
                    text.addAll(pitch.text(this));
                }

                spectrogram.paintBar(this, g);
                spectrogram.paintCursor(this, g, Color.WHITE, text);
            }
        };

        emptyView = new JPanel();

        view.setLayout(new BoxLayout(view, BoxLayout.Y_AXIS));
        view.add(waveformView);

        ChangeListener updateView = e -> {
            var waveform = waveformToggle.isSelected();
            var spectrogram = spectrogramToggle.isSelected();
            var pitch = pitchToggle.isSelected();

            view.removeAll();

            if (waveform)
                view.add(waveformView);
            if (spectrogram || pitch)
                view.add(spectrogramView);
            if (!waveform && !spectrogram && !pitch)
                view.add(emptyView);

            pack();
            view.repaint();
        };
        waveformToggle.setSelected(true);
        waveformToggle.addChangeListener(updateView);
        spectrogramToggle.addChangeListener(updateView);
        pitchToggle.addChangeListener(updateView);

        waveform.addListeners(waveformView);
        spectrogram.addListeners(spectrogramView);
        pitch.addListeners(spectrogramView);

        zoom.setValue((int)SSMath.project(
            ZOOM_MAX, ZOOM_MIN,
            WINDOW_MIN.absolute(), WINDOW_MAX.absolute(),
            WINDOW_DEFAULT.absolute()));

        zoom.addChangeListener(e -> {
            var ns = (long)SSMath.project(
                WINDOW_MAX.absolute(), WINDOW_MIN.absolute(),
                ZOOM_MIN, ZOOM_MAX,
                zoom.getValue());
            var page = Duration.ns(ns);
            System.out.println("Page: " + page.as(TimeUnit.SECOND));
            waveform.page(page);
            spectrogram.page(page);
            pitch.page(page);
            view.repaint();
        });

        largerSize = new Dimension(1400, 450);
        smallerSize = new Dimension(800, 225);
        setPreferredSize(largerSize);

        resize.addActionListener(l -> {
            SS3Singleton.getMediaPlayer().audios()
                .forEach(a -> ((AudioView) a.view()).toggle());
        });
        resize.setToolTipText("Smaller Size");

        build();
    }

    public void toggle() {
        var size = resize.state() ? largerSize : smallerSize;
        setPreferredSize(size);
        setSize(size);
        resize.toggle();

        resize.setToolTipText(resize.state() ? "Larger Size" : "Smaller Size");
        pack();
    }
}
