package edu.bu.signstream.media.ui.painter;

import edu.bu.signstream.common.SSMath;
import edu.bu.signstream.time.Duration;
import edu.bu.signstream.time.TimeRange;
import edu.bu.signstream.media.data.FFT;
import edu.bu.signstream.media.data.Samples;
import edu.bu.signstream.media.data.Window;
import edu.bu.signstream.media.ui.color.ColorMap;

import java.awt.*;
import java.util.List;

public class SpectrogramPainter extends SamplesPainter {
    // FFT window function, size, and stride
    private static final Window windowFn = Window.GAUSSIAN;
    private static final int windowSize = 2048;
    private static final int stride = 256;

    // Display frequency range
    private double f0 = 0;
    private double f1 = 5000.0;

    private final FFT[] data;

    public SpectrogramPainter(Samples samples, ColorMap colors, TimeRange range, Duration page) {
        super(samples, colors, range, page);

        var start = System.currentTimeMillis();

        // Compute the FFTs for each window
        data = new FFT[(int)Math.ceil((double)samples.length / stride)];
        for (int w = 0; w < data.length; w++) {
            var i = w * stride;
            var j = Math.min(w * stride + windowSize, samples.length);
            data[w] = samples.fftForward(windowFn, i, j);
        }

        System.out.println("P: Calculated spectrogram FFTs in " + (System.currentTimeMillis() - start) + " ms");
    }

    public void f0(double freq) { this.f0 = freq; }
    public void f1(double freq) { this.f1 = freq; }

    @Override
    protected void paint(Graphics g, int w, int h, int s0, int s1) {
        // Get start and end bins based on the frequency range
        var bin0 = data[0].bin(f0);
        var bin1 = data[0].bin(f1);

        for (int i = 0; i < w; i++) {
            // Get the sample range for this column
            int j0 = (int)SSMath.project(s0, s1, 0, w, i);
            int j1 = (int)SSMath.project(s0, s1, 0, w, i + 1);

            // Get the window range for this column
            var w0 = j0 / stride;
            var w1 = Math.max(1, Math.max(w0 + 1, j1 / stride));

            if (w0 >= data.length || w1 >= data.length)
                continue;

            // Get the max amplitude in the frequency range over all
            // windows for this column
            var max = Double.MIN_VALUE;
            for (int j = w0; j <= w1; j++)
                max = Math.max(max, data[j].max(bin0, bin1));

            for (int j = 0; j < h; j++) {
                // Get the logarithmic frequency bin corresponding to the height
                var bin = (int)SSMath.lerp(bin0, bin1, (double)j / h);

                // Get the average amplitude for the frequency bin over each window,
                // normalized to the ceiling
                var s = 0.0;
                for (int k = w0; k <= w1; k++)
                    s += data[k].get(bin);
                s = (s / Math.max(1, w1 - w0)) / max;

                // Plot a color mapped point based on the normalized amplitude
                g.setColor(colors.map(s));
                g.drawLine(i, h - j, i, h - j);
            }
        }
    }

    @Override
    public List<String> text(int w, int h, int mx, int my) {
        var text = super.text(w, h, mx, my);
        var freq = SSMath.project(f1, f0, 0, h, my);
        text.add(String.format("Spectrogram: %,.2f Hz", freq));
        return text;
    }
}
