package edu.bu.signstream.time;

import edu.bu.signstream.common.util.Util;
import edu.bu.signstream.grepresentation.requestsHandler.SS3Singleton;
import edu.bu.signstream.media.gstreamer.GstMediaPlayer;

public class TimeRange {
    private final String id;
    private final String name;
	private final TimeRange parent;
	private final TimeUnit unit;

	private Duration offset;
	private Duration duration;

	private long step;

    /**
     * `TimeRange`s should only be constructed by `TimeRange.subrange`, or from the `GlobalTimeRange`.
     */
	TimeRange(String id, String name, TimeRange parent, TimeUnit unit, Duration offset, Duration duration, long step) {
        this.id = id;
	    this.name = name;
	    this.parent = parent;
        this.unit = unit;
        this.offset = parent == null ? offset : offset.convert(parent.unit());
	    this.duration = duration.convert(unit);
	    this.step = step;
    }
    TimeRange(String name, TimeRange parent, TimeUnit unit, Duration offset, Duration duration, long step) {
	    this(Util.uuid(), name, parent, unit, offset, duration, step);
    }

    public String id() { return id; }
    public String name() { return name; }
	public TimeRange parent() { return parent; }
	public TimeUnit unit() { return unit; }
    public Duration offset() { return offset; }
    public Duration duration() { return duration; }
    public long step() { return step; }

    public void offset(Duration offset) { this.offset = offset.convert(parent().unit()); }
    public void duration(Duration duration) { this.duration = duration; }
    public void step(long step) { this.step = step; }

    /**
     * Zero Time in this TimeRange.
     */
    public Time zero() { return new Time(this, 0, unit); }

    /**
     * Starting point of the TimeRange in the ABSOLUTE TimeRange.
     */
    public Time start() { return parent().start().add(offset()); }
    public Time end() { return start().add(duration()); }
    public Interval interval() { return start().to(end()); }


    /**
     * Construct a child `TimeRange` with some offset and a new unit.
     */
    public TimeRange subrange(String id, String name, TimeUnit unit, Duration offset, Duration duration, long step) {
        return new TimeRange(id, name, this, unit, offset, duration, step);
    }
    public TimeRange subrange(String name, TimeUnit unit, Duration offset, Duration duration, long step) {
        return new TimeRange(name, this, unit, offset, duration, step);
    }


    /**
     * Construct a `Time` in the range with a given time from the start of the range.
     *
     * JACK TODO: Track children for updates?
     */
    public Time time(long time, TimeUnit unit, Rounding r) {
        return new Time(this, time, unit).convert(unit(), r);
    }
    public Time time(long time, TimeUnit unit) {
        return new Time(this, time, unit).convert(unit());
    }

    @Override
    public boolean equals(Object obj) {
        return obj instanceof TimeRange && id.equals(((TimeRange)obj).id);
    }

    @Override
    public String toString() { return name; }

    public static final String ABSOLUTE_ID = "0";
    public static final String VIDEO_ID = "1";
    public static final String AUDIO_ID = "2";

    /**
     * The absolute time range, measured in nanoseconds, is the base point off of which all other
     * time ranges are constructed. Zero in this range can be thought of as "absolute zero".
     */
    public static final TimeRange ABSOLUTE = new TimeRange(ABSOLUTE_ID, "Absolute", null, TimeUnit.NANOSECOND, Duration.ZERO, Duration.ZERO, 10_000_000) {
        @Override
        public void offset(Duration offset) { throw new IllegalStateException("Tried to offset ABSOLUTE TimeRage!"); }

        @Override
        public Time start() { return new Time(this, 0, TimeUnit.NANOSECOND); }

        @Override
        public Duration duration() {
            var player = SS3Singleton.getMediaPlayer();

            if (!((GstMediaPlayer)player).prerolled())
                return Duration.ZERO;
            else
                return player.duration();
        }

        @Override
        public void duration(Duration duration) { throw new IllegalStateException("Tried to set duration of ABSOLUTE TimeRage!"); }
    };

    public static final TimeRange FRAME = new TimeRange(VIDEO_ID, "Frame", ABSOLUTE, TimeUnit.FRAME, Duration.ZERO, Duration.ZERO, 1) {
        @Override
        public void offset(Duration offset) { throw new IllegalStateException("Tried to offset FRAME TimeRage!"); }

        @Override
        public Time start() {
            return new Time(this, 0, TimeUnit.FRAME);
        }

        @Override
        public Duration duration() {
            var player = SS3Singleton.getMediaPlayer();

            if (!((GstMediaPlayer)player).prerolled())
                return Duration.ZERO;
            else
                return player.duration().convert(TimeUnit.FRAME);
        }

        @Override
        public void duration(Duration duration) { throw new IllegalStateException("Tried to set duration of FRAME TimeRage!"); }
    };
}
